/* Pavel Surynek, 2008 - 2010 */
/* pavel.surynek@mff.cuni.cz */
/* bibox_main.cpp */

/*
   The BIBOX Algorithm

   This is an implementation of the multi-robot path planning algorithm for biconnected environments with
   at least two free places. The code is not commented however it is simple enough to be understood
   without comments.

   The problem is given as an undirected graph, where each vertex of the graph is either free or
   occupied by a robot. A robot can move to a neighboring unoccupied vertex supposed that no other
   robot is simultaneously entering the target vertex. The task is to move robots preserving this
   condition to their final positions.

   This program is an implementation of the algorithm for solving the problem where the graph is
   biconnected and threre are at least two unoccupied vertices in the graph. The program runs in
   polynomial time with respect to the size of the input graph. More precisely it runs in O(|V|^3)
   where G=(V,E) is the input biconnected graph.
 */

#include <string.h>

#include "slider.h"
#include "bibox.h"
#include "obox.h"

typedef enum {
	BIBOX,
	BIBOX_MIT,
	BIBOX_THETA_2,
	BIBOX_THETA_3,
	MIT
} METHOD_TYPE;

typedef enum {
	RAW,
	INVERSE,
	REDUNDANT,
	LONG
} SHORTENING_TYPE;

const int MAX_SEED_TRIES = 128;

void test_grid_shortening(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	int Size_X;
	int Size_Y;
	int Robots;

	int Seed;

	for (int SizeXY = 4; SizeXY <= 20; ++SizeXY)
	{
		Size_X = Size_Y = SizeXY;
		
		Seed = 121;

		Robots = Size_X * Size_Y - 1;

		char prefix[1024];
		char suffix[1024];

		switch (method_type)
		{
		case BIBOX:
		{
			sprintf(prefix, "bibox_");
			break;
		}
		case BIBOX_MIT:
		{
			sprintf(prefix, "bibox_mit_");
			break;
		}
		case BIBOX_THETA_2:
		{
			sprintf(prefix, "bibox_theta_2_");
			break;
		}
		case BIBOX_THETA_3:
		{
			sprintf(prefix, "bibox_theta_3_");
			break;
		}
		case MIT:
		{
			sprintf(prefix, "mit_");
			break;
		}
		default:
		{
			break;
		}
		}

		switch (shortening_type)
		{
		case RAW:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_raw.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		case INVERSE:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_inverse.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		case REDUNDANT:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_redundant.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		case LONG:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_long.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		default:
		{
			break;
		}
		}

		if (method_type == BIBOX)
		{
		    --Robots;
		}

		char filename[1024];
		sprintf(filename, "%s%s", prefix, suffix);

		FILE *fw;
		fw = fopen(filename, "w");

		for (int r = 1; r <= Robots; ++r)
		{
			clock_t begin_1, end_1;
			time_t sec_begin_1, sec_end_1;
			clock_t begin_2, end_2;
			time_t sec_begin_2, sec_end_2;
		
			GRAPH grid1, grid2, grid3;
			generate_grid(Size_X, Size_Y, &grid1);
			generate_grid(Size_X, Size_Y, &grid2);
			generate_grid(Size_X, Size_Y, &grid3);
			
			generate_grid_random(r, Seed, &grid1);
			generate_grid_random(r, Seed, &grid2);
			generate_grid_random(r, Seed, &grid3);

			begin_1 = clock();
			sec_begin_1 = time(&sec_begin_1);

			SOLUTION_DATABASE database;
			FINE_SOLUTION solution, filtered_solution;

			printf("Solving ...\n");
			fprintf(fw, "Robots:%d\n", r);
			fprintf(fw, "Size:%d x %d\n", Size_X, Size_Y);

			bool solved;

			switch (method_type)
			{
			case BIBOX:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);
				break;
			}
			case BIBOX_MIT:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, true);
				break;
			}
			case BIBOX_THETA_2:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);
				break;
			}
			case BIBOX_THETA_3:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);
				break;
			}
			case MIT:
			{
				int solved_cycles = 0;
				solved = solve_slider_obox_theta(&grid1, database, solution, false, false, solved_cycles);
				break;
			}
			default:
			{
				solved = false;
				break;
			}
			}
// Bibox_theta 2
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);

// Bibox_theta 3
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);

// Bibox_mit 
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, true);

// Bibox
//			solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);

// Mit
//			solved = solve_slider_obox_theta(&grid1, database, solution, false, false);

			end_1 = clock();
			sec_end_1 = time(&sec_end_1);
			begin_2 = clock();
			sec_begin_2 = time(&sec_begin_2);

			filter_fine_solution(&grid2, solution, filtered_solution);

			switch (shortening_type)
			{
			case RAW:
			{
				break;
			}
			case INVERSE:
			{
				erase_trivial_inverses(filtered_solution);
				break;
			}
			case REDUNDANT:
			{
				erase_independent_inverses(filtered_solution);
				break;
			}
			case LONG:
			{
				shorten_solution(filtered_solution, &grid3);
				break;
			}
			default:
			{
				break;
			}
			}

//			erase_trivial_inverses(filtered_solution);
//			erase_independent_inverses(filtered_solution);
//			shorten_solution(filtered_solution, &grid3);
//			execute_fine_solution(filtered_solution, &grid3);

			parallelize_fine_solution_multirobotic_FAST(filtered_solution);

			end_2 = clock();
			sec_end_2 = time(&sec_end_2);

			fprintf(fw, "Solution size:%d\n", solution.size());
			fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());
			fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
			fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
			fprint_step_parallelism(fw, filtered_solution);
		
			fprintf(fw, "Solving time:%.3f,%d\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC, sec_end_1 - sec_begin_1);
			fprintf(fw, "Filtering time:%.3f,%d\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC, sec_end_2 - sec_begin_2);

			fflush(fw);
		}
	}
}


void test_grid_shortening_BIG(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	int Size_X;
	int Size_Y;
	int Robots;

	int Seed;

	for (int SizeXY = 48; SizeXY <= 80; SizeXY += 8)
	{
		Size_X = Size_Y = SizeXY;
		
		Seed = 121;

		Robots = Size_X * Size_Y - 1;

		char prefix[1024];
		char suffix[1024];

		switch (method_type)
		{
		case BIBOX:
		{
			sprintf(prefix, "bibox_");
			break;
		}
		case BIBOX_MIT:
		{
			sprintf(prefix, "bibox_mit_");
			break;
		}
		case BIBOX_THETA_2:
		{
			sprintf(prefix, "bibox_theta_2_");
			break;
		}
		case BIBOX_THETA_3:
		{
			sprintf(prefix, "bibox_theta_3_");
			break;
		}
		case MIT:
		{
			sprintf(prefix, "mit_");
			break;
		}
		default:
		{
			break;
		}
		}

		switch (shortening_type)
		{
		case RAW:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_raw.BIG.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		case INVERSE:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_inverse.BIG.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		case REDUNDANT:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_redundant.BIG.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		case LONG:
		{
			sprintf(suffix, "grid_|%s|_%dx%d_long.BIG.txt", SPECIFICATION, Size_X, Size_Y);
			break;
		}
		default:
		{
			break;
		}
		}

		if (method_type == BIBOX)
		{
		    --Robots;
		}

		char filename[1024];
		sprintf(filename, "%s%s", prefix, suffix);

		FILE *fw;
		fw = fopen(filename, "w");

		for (int r = 1; r <= Robots; ++r)
		{
			clock_t begin_1, end_1;
			time_t sec_begin_1, sec_end_1;
			clock_t begin_2, end_2;
			time_t sec_begin_2, sec_end_2;
		
			GRAPH grid1, grid2, grid3;
			generate_grid(Size_X, Size_Y, &grid1);
			generate_grid(Size_X, Size_Y, &grid2);
			generate_grid(Size_X, Size_Y, &grid3);
			
			generate_grid_random(r, Seed, &grid1);
			generate_grid_random(r, Seed, &grid2);
			generate_grid_random(r, Seed, &grid3);

			begin_1 = clock();
			sec_begin_1 = time(&sec_begin_1);

			SOLUTION_DATABASE database;
			FINE_SOLUTION solution, filtered_solution;

			printf("Solving ...\n");
			fprintf(fw, "Robots:%d\n", r);
			fprintf(fw, "Size:%d x %d\n", Size_X, Size_Y);

			bool solved;

			switch (method_type)
			{
			case BIBOX:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);
				break;
			}
			case BIBOX_MIT:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, true);
				break;
			}
			case BIBOX_THETA_2:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);
				break;
			}
			case BIBOX_THETA_3:
			{
				solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);
				break;
			}
			case MIT:
			{
				int solved_cycles = 0;
				solved = solve_slider_obox_theta(&grid1, database, solution, false, false, solved_cycles);
				break;
			}
			default:
			{
				solved = false;
				break;
			}
			}
// Bibox_theta 2
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);

// Bibox_theta 3
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);

// Bibox_mit 
//			solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, true);

// Bibox
//			solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);

// Mit
//			solved = solve_slider_obox_theta(&grid1, database, solution, false, false);

			end_1 = clock();
			sec_end_1 = time(&sec_end_1);
			begin_2 = clock();
			sec_begin_2 = time(&sec_begin_2);

			filter_fine_solution(&grid2, solution, filtered_solution);

			switch (shortening_type)
			{
			case RAW:
			{
				break;
			}
			case INVERSE:
			{
				erase_trivial_inverses(filtered_solution);
				break;
			}
			case REDUNDANT:
			{
				erase_independent_inverses(filtered_solution);
				break;
			}
			case LONG:
			{
				shorten_solution(filtered_solution, &grid3);
				break;
			}
			default:
			{
				break;
			}
			}

//			erase_trivial_inverses(filtered_solution);
//			erase_independent_inverses(filtered_solution);
//			shorten_solution(filtered_solution, &grid3);
//			execute_fine_solution(filtered_solution, &grid3);

			parallelize_fine_solution_multirobotic_FAST(filtered_solution);

			end_2 = clock();
			sec_end_2 = time(&sec_end_2);

			fprintf(fw, "Solution size:%d\n", solution.size());
			fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());
			fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
			fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
			fprint_step_parallelism(fw, filtered_solution);
		
			fprintf(fw, "Solving time:%.3f,%d\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC, sec_end_1 - sec_begin_1);
			fprintf(fw, "Filtering time:%.3f,%d\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC, sec_end_2 - sec_begin_2);

			fflush(fw);
		}
	}
}


void test_exchange_grids(METHOD_TYPE UNUSED(method_type = BIBOX_THETA_2), SHORTENING_TYPE UNUSED(shortening_type = RAW))
{
	clock_t begin, end;
/*
	GRAPH biconnected1, biconnected2;

	srand(123);
	generate_biconnected(5, 5, 10, 2, &biconnected1);

	srand(123);
	generate_biconnected(5, 5, 10, 2, &biconnected2);

	biconnected1.print();
*/

/* Grid 1 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(10, 6, &grid1);
	generate_grid(10, 6, &grid2);
	generate_grid(10, 6, &grid3);

	generate_grid_exchange(10, 6, 2, 0, 3, 5, 8, 0, &grid1);
	generate_grid_exchange(10, 6, 2, 0, 3, 5, 8, 0, &grid2);
	generate_grid_exchange(10, 6, 2, 0, 3, 5, 8, 0, &grid3);

	grid1.print();
*/

/* Grid 2 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(16, 10, &grid1);
	generate_grid(16, 10, &grid2);
	generate_grid(16, 10, &grid3);

	generate_grid_exchange(16, 10, 2, 0, 3, 9, 14, 0, &grid1);
	generate_grid_exchange(16, 10, 2, 0, 3, 9, 14, 0, &grid2);
	generate_grid_exchange(16, 10, 2, 0, 3, 9, 14, 0, &grid3);

	grid1.print();
*/

/* Grid 3 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(16, 10, &grid1);
	generate_grid(16, 10, &grid2);
	generate_grid(16, 10, &grid3);

	generate_grid_exchange(16, 10, 2, 0, 4, 9, 13, 0, &grid1);
	generate_grid_exchange(16, 10, 2, 0, 4, 9, 13, 0, &grid2);
	generate_grid_exchange(16, 10, 2, 0, 4, 9, 13, 0, &grid3);

	grid1.print();
*/

/* Grid 4 */

	GRAPH grid1, grid2, grid3;
	generate_grid(20, 16, &grid1);
	generate_grid(20, 16, &grid2);
	generate_grid(20, 16, &grid3);

	generate_grid_exchange(20, 16, 2, 0, 4, 15, 16, 0, &grid1);
	generate_grid_exchange(20, 16, 2, 0, 4, 15, 16, 0, &grid2);
	generate_grid_exchange(20, 16, 2, 0, 4, 15, 16, 0, &grid3);

	grid1.print();


/* Grid 5 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(20, 16, &grid1);
	generate_grid(20, 16, &grid2);
	generate_grid(20, 16, &grid3);

	generate_grid_exchange(20, 16, 2, 0, 6, 15, 14, 0, &grid1);
	generate_grid_exchange(20, 16, 2, 0, 6, 15, 14, 0, &grid2);
	generate_grid_exchange(20, 16, 2, 0, 6, 15, 14, 0, &grid3);

	grid1.print();
*/
/* Grid 6 */
/*
	GRAPH grid1, grid2, grid3;
	generate_grid(10, 6, &grid1);
	generate_grid(10, 6, &grid2);
	generate_grid(10, 6, &grid3);

	generate_grid_exchange(10, 6, 2, 0, 5, 5, 6, 0, &grid1);
	generate_grid_exchange(10, 6, 2, 0, 5, 5, 6, 0, &grid2);
	generate_grid_exchange(10, 6, 2, 0, 5, 5, 6, 0, &grid3);

	grid1.print();
*/
/*	
	printf("Solving ...\n");

	bool solved;
	solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);
*/
/*
	grid2.print();
	grid1.print();
*/
	begin = clock();   

	SOLUTION_DATABASE database;
	FINE_SOLUTION solution, filtered_solution;

	printf("Solving ...\n");

// Bibox_theta 2
//	solved = solve_slider_bibox_theta(&grid1, database, solution, true, false, false);

// Bibox_theta 3
//	solved = solve_slider_bibox_theta(&grid1, database, solution, true, true, false);

// Bibox_mit 
//	solved = solve_slider_bibox_theta(&biconnected1, database, solution, true, false, true);

// Bibox
//	solved = solve_slider_bibox_theta(&biconnected1, database, solution, false, false, false);
//	solved = solve_slider_bibox_theta(&grid1, database, solution, false, false, false);

//	printf("Finished ...\n");

//	filter_fine_solution(&biconnected2, solution, filtered_solution);

	filter_fine_solution(&grid2, solution, filtered_solution);
//	erase_trivial_inverses(filtered_solution);
//	erase_independent_inverses(filtered_solution);
	shorten_solution(filtered_solution, &grid3);
	parallelize_fine_solution_multirobotic_FAST(filtered_solution);

	print_fine_solution(filtered_solution);

	printf("Solution size:%d\n", solution.size());
	printf("Filtered solution size:%d\n", filtered_solution.size());
	printf("Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
	printf("Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
	print_step_parallelism(filtered_solution);

	end = clock();
	printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);

/*
	int dur = calc_solution_duration(filtered_solution);
	for (int k = 0; k <= dur; ++k)
	{
		printf("Step:%d\n", k);
		show_state_in_grid(30, 20, &grid3, k, filtered_solution);
		printf("\n");
	}
*/
}


void test_initial(void)
{
//	exit(0);
/*
	g.m_cycles.push_back(CYCLE());
	CYCLE *cycle = &g.m_cycles.back();

	cycle->m_x_id = 3; // Same 1
	cycle->m_y_id = 5; // Same 1 + 2

	g.add_vertex(1, 0, 6, 6);
	cycle->m_vertices.push_back(1);
	cycle->m_loop.push_back(1);
	g.add_vertex(2, 0, 5, 5);
	cycle->m_vertices.push_back(2);
	cycle->m_loop.push_back(2);
	g.add_vertex(3, 0, 1, 3);  // Same 1
	cycle->m_vertices.push_back(3);
	cycle->m_loop.push_back(3);
	g.add_vertex(4, 0, 2, 4);
	cycle->m_vertices.push_back(4);
	cycle->m_loop.push_back(4);
	g.add_vertex(5, 0, 8, 11);
	cycle->m_vertices.push_back(5);
	cycle->m_loop.push_back(5);
	g.add_vertex(6, 0, 7, 12);
	cycle->m_vertices.push_back(6);
	cycle->m_loop.push_back(6);
	g.add_vertex(7, 0, -1, 7);
	cycle->m_vertices.push_back(7);
	cycle->m_loop.push_back(7);
	g.add_vertex(8, 0, 0, 8);
	cycle->m_vertices.push_back(8);
	cycle->m_loop.push_back(8);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 4;
	cycle->m_y_id = 8;
	g.add_vertex(9, 1, 3, 1);
	cycle->m_vertices.push_back(9);
	cycle->m_loop.push_back(9);
	g.add_vertex(10, 1, 4, 2);
	cycle->m_vertices.push_back(10);
	cycle->m_loop.push_back(10);
	g.add_vertex(11, 1, 11, 0);
	cycle->m_vertices.push_back(11);
	cycle->m_loop.push_back(11);
	g.add_vertex(12, 1, 12, -1);
	cycle->m_vertices.push_back(12);
	cycle->m_loop.push_back(12);

	cycle->m_vertices.push_back(8);
	cycle->m_vertices.push_back(4);

	g.add_edge(1, 2, 0);
	g.add_edge(2, 3, 0);
	g.add_edge(3, 4, 0);
	g.add_edge(4, 5, 0);
	g.add_edge(5, 6, 0);
	g.add_edge(6, 7, 0);
	g.add_edge(7, 8, 0);
	g.add_edge(8, 1, 0);

	g.add_edge(4, 9, 1);
	g.add_edge(9, 10, 1);
	g.add_edge(10, 11, 1);
	g.add_edge(11, 12, 1);
	g.add_edge(12, 8, 1);
*/
/*
	g.m_cycles.push_back(CYCLE());
	CYCLE *cycle = &g.m_cycles.back();

	cycle->m_x_id = 2;
	cycle->m_y_id = 4; // Same 1

	g.add_vertex(1, 0, -1, 5);
	cycle->m_vertices.push_back(1);
	cycle->m_loop.push_back(1);
	g.add_vertex(2, 0, 0, 15);
	cycle->m_vertices.push_back(2);
	cycle->m_loop.push_back(2);
	g.add_vertex(3, 0, 1, 2);
	cycle->m_vertices.push_back(3);
	cycle->m_loop.push_back(3);
	g.add_vertex(4, 0, 2, 4); // Same 1
	cycle->m_vertices.push_back(4);
	cycle->m_loop.push_back(4);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 3;
	cycle->m_y_id = 4;
	g.add_vertex(5, 1, 3, 1);
	cycle->m_vertices.push_back(5);
	cycle->m_loop.push_back(5);
	g.add_vertex(6, 1, 4, 3);
	cycle->m_vertices.push_back(6);
	cycle->m_loop.push_back(6);

	cycle->m_vertices.push_back(4);
	cycle->m_vertices.push_back(3);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 5;
	cycle->m_y_id = 6;
	g.add_vertex(7, 2, 5, 6);
	cycle->m_vertices.push_back(7);
	cycle->m_loop.push_back(7);
	g.add_vertex(8, 2, 6, 13);
	cycle->m_vertices.push_back(8);
	cycle->m_loop.push_back(8);

	cycle->m_vertices.push_back(6);
	cycle->m_vertices.push_back(5);

	g.m_cycles.push_back(CYCLE());
	cycle = &g.m_cycles.back();

	cycle->m_x_id = 1;
	cycle->m_y_id = 8;
	g.add_vertex(9, 3, 15, 9);
	cycle->m_vertices.push_back(9);
	cycle->m_loop.push_back(9);
	g.add_vertex(10, 3, 14, 12);
	cycle->m_vertices.push_back(10);
	cycle->m_loop.push_back(10);
	g.add_vertex(11, 3, 13, 11);
	cycle->m_vertices.push_back(11);
	cycle->m_loop.push_back(11);
	g.add_vertex(12, 3, 12, 10);
	cycle->m_vertices.push_back(12);
	cycle->m_loop.push_back(12);
	g.add_vertex(13, 3, 11, 14);
	cycle->m_vertices.push_back(13);
	cycle->m_loop.push_back(13);
	g.add_vertex(14, 3, 10, 0);
	cycle->m_vertices.push_back(14);
	cycle->m_loop.push_back(14);
	g.add_vertex(15, 3, 9, -1);
	cycle->m_vertices.push_back(15);
	cycle->m_loop.push_back(15);

	cycle->m_vertices.push_back(8);
	cycle->m_vertices.push_back(6);
	cycle->m_vertices.push_back(4);
	cycle->m_vertices.push_back(1);

	g.add_edge(1, 2, 0);
	g.add_edge(2, 3, 0);
	g.add_edge(3, 4, 0);
	g.add_edge(4, 1, 0);
	g.add_edge(3, 5, 1);
	g.add_edge(5, 6, 1);
	g.add_edge(6, 4, 1);
	g.add_edge(5, 7, 2);
	g.add_edge(7, 8, 2);
	g.add_edge(8, 6, 2);
	g.add_edge(1, 9, 3);
	g.add_edge(9, 10, 3);
	g.add_edge(10, 11, 3);
	g.add_edge(11, 12, 3);
	g.add_edge(12, 13, 3);
	g.add_edge(13, 14, 3);
	g.add_edge(14, 15, 3);
	g.add_edge(15, 8, 3);
*/

//	FILE *s;
//	s = fopen("large_problems.txt", "w");
/*
	FILE *b_lpg, *b_sgplan, *b_stan, *b_ipp, *b_max, *b_sat;
	b_lpg = fopen("batch_LPG_lenght=3_free=90.bat", "w");
	b_sgplan = fopen("batch_SGPLAN_lenght=3_free=90.bat", "w");
	b_stan = fopen("batch_STAN_lenght=3_free=90.bat", "w");
	b_ipp = fopen("batch_IPP_lenght=3_free=90.bat", "w");
	b_max = fopen("batch_MAX_lenght=3_free=90.bat", "w");
	b_sat = fopen("batch_SAT_lenght=3_free=90.bat", "w");
*/
}


void test_biconnected_shortening(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	clock_t begin_1, end_1;
	time_t sec_begin_1, sec_end_1;
	clock_t begin_2, end_2;
	time_t sec_begin_2, sec_end_2;

	int Seed;
	int init;
	int iter;
	int length;

	for (length = 2; length <= 16; ++length)
	{
		int free = 2;
		int ID = 0;

		switch (length)
		{
		case 2:
		{
			Seed = 10;
			init = 5;
			iter = 43;
			length = 2;
			// |V|=91
			break;
		}
		case 3:
		{
			Seed = 17;
			init = 5;
			iter = 42;
			length = 3;
			// |V|=89	
			break;
		}
		case 4:
		{
			Seed = 16;
			init = 7;
			iter = 38;
			length = 4;
			// |V|=91	
			break;
		}
		case 5:
		{
			Seed = 1013;
			init = 7;
			iter = 32;
			length = 5;
			// |V|=89	
			break;
		}
		case 6:
		{
			Seed = 1214;
			init = 7;
			iter = 30;
			length = 6;
			// |V|=89	
			break;
		}
		case 7:
		{
			Seed = 1227;
			init = 7;
			iter = 26;
			length = 7;
			// |V|=90	
			break;
		}
		case 8:
		{
			Seed = 1254;
			init = 7;
			iter = 22;
			length = 8;
			// |V|=91	
			break;
		}
		case 9:
		{
			Seed = 1261;
			init = 7;
			iter = 20;
			length = 9;
			// |V|=90
			break;
		}
		case 10:
		{
			Seed = 1267;
			init = 7;
			iter = 18;
			length = 10;
			// |V|=90	
			break;
		}
		case 11:
		{
			Seed = 1279;
			init = 7;
			iter = 16;
			length = 11;
			// |V|=91	
			break;
		}
		case 12:
		{
			Seed = 1301;
			init = 7;
			iter = 14;
			length = 12;
			// |V|=91	
			break;
		}
		case 13:
		{
			Seed = 1306;
			init = 7;
			iter = 13;
			length = 13;
			// |V|=91	
			break;
		}
		case 14:
		{
			Seed = 1317;
			init = 9;
			iter = 12;
			length = 14;
			// |V|=91	
			break;
		}
		case 15:
		{		
			Seed = 1323;
			init = 9;
			iter = 10;
			length = 15;
			// |V|=91	
			break;
		}
		case 16:
		{
			Seed = 1342;
			init = 11;
			iter = 9;
			length = 16;
			// |V|=89	
			break;
		}
		default:
		{
			break;
		}
		}

		char prefix[1024];
		char suffix[1024];
		char filename[1024];

		switch (method_type)
		{
		case BIBOX:
		{
			sprintf(prefix, "bibox_");
			break;
		}
		case BIBOX_MIT:
		{
			sprintf(prefix, "bibox_mit_");
			break;
		}
		case BIBOX_THETA_2:
		{
			sprintf(prefix, "bibox_theta_2_");
			break;
		}
		case BIBOX_THETA_3:
		{
			sprintf(prefix, "bibox_theta_3_");
			break;
		}
		case MIT:
		{
			sprintf(prefix, "mit_");
			break;
		}
		default:
		{
			prefix[0] = '\0';
			break;
		}
		}

		switch (shortening_type)
		{
		case RAW:
		{
			sprintf(suffix, "biconnected_%s_%d-%d-%d_raw.txt", SPECIFICATION, init, iter, length);
			break;
		}
		case INVERSE:
		{
			sprintf(suffix, "biconnected_%s_%d-%d-%d_inverse.txt", SPECIFICATION, init, iter, length);
			break;
		}
		case REDUNDANT:
		{
			sprintf(suffix, "biconnected_%s_%d-%d-%d_redundant.txt", SPECIFICATION, init, iter, length);
			break;
		}
		case LONG:
		{
			sprintf(suffix, "biconnected_%s_%d-%d-%d_long.txt", SPECIFICATION, init, iter, length);
			break;
		}
		default:
		{
			prefix[0] = '\0';
			break;
		}
		}

		FILE *fw;
		sprintf(filename, "%s%s", prefix, suffix);

		fw = fopen(filename, "w");

		if (fw == NULL)
		{
			printf("Cannot open %s. Terminating...\n", filename);
			exit(-1);
		}

		int start_free;

		if (method_type == BIBOX)
		{
			start_free = 2;
		}
		else
		{
			start_free = 1;
		}

		printf("Search for best seed... ");
		fflush(NULL);

		int best_Seed = Seed;
		int best_cycles = 0;
		
		for (int s = 0; s < MAX_SEED_TRIES; ++s)
		{
			GRAPH h;
			SOLUTION_DATABASE database;
			FINE_SOLUTION solution;

			srandom(Seed);
			generate_biconnected(init, iter, length, 2, &h);

			int cycles = 0;
			bool solved = solve_slider_obox_theta(&h, database, solution, false, false, cycles);

			if (solved)
			{
				best_Seed = Seed;
				break;
			}
			else
			{
				if (cycles > best_cycles)
				{
					printf("%d ", cycles);
					fflush(NULL);

					best_Seed = Seed;
					best_cycles = cycles;
				}
			}
			
			Seed = random();
		}

		printf("\n");
		Seed = best_Seed;
	       
		for (free = start_free; /****/; ++free)
		{
			fprintf(fw, "ID:%d (%d,%d,%d)\n", ID, iter, length, free);
			++ID;
			
			SOLUTION_DATABASE database;
			FINE_SOLUTION solution, filtered_solution;
			
			MOVE_COUNT = 0;
			STEP_COUNT = 0;
			GRAPH g, h, ch, hv;
			
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &g);
			
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &h);
			
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &ch);
			
			srandom(Seed);
			generate_biconnected(init, iter, length, free, &hv);
		
			fprintf(fw, "Graph |V|:%d\n", g.m_vertices.size());
			
			if (h.m_vertices.size() > (unsigned int)free)
			{
				begin_1 = clock();
				sec_begin_1 = time(&sec_begin_1);
				
				bool solved;
				
				switch (method_type)
				{
				case BIBOX:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
					break;
				}
				case BIBOX_MIT:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, true, false, true);
					break;
				}
				case BIBOX_THETA_2:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
					break;
				}
				case BIBOX_THETA_3:
				{
					solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
					break;
				}
				case MIT:
				{
					int solved_cycles = 0;
					solved = solve_slider_obox_theta(&h, database, solution, false, false, solved_cycles);
					break;
				}
				default:
				{
					solved = false;
					break;
				}
				}
// Bibox_theta 2	
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
// Bibox_theta 3
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
// Bibox
//			bool solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
			
				end_1 = clock();
				sec_end_1 = time(&sec_end_1);
				begin_2 = clock();
				sec_begin_2 = time(&sec_begin_2);
				
				filter_fine_solution(&ch, solution, filtered_solution);
				
				switch (shortening_type)
				{
				case RAW:
				{
					break;
				}
				case INVERSE:
				{
					erase_trivial_inverses(filtered_solution);
					break;
				}
				case REDUNDANT:
				{
					erase_independent_inverses(filtered_solution);
					break;
				}
				case LONG:
				{
					shorten_solution(filtered_solution, &hv);
					break;
				}
				default:
				{
					break;
				}
				}
			
//		erase_trivial_inverses(filtered_solution);
//		erase_complex_inverses(filtered_solution);
//		erase_long_inverses(filtered_solution);
//		erase_independent_inverses(filtered_solution);
//		shorten_solution(filtered_solution, &hv);
			
				parallelize_fine_solution_multirobotic_FAST(filtered_solution);
				
				end_2 = clock();
				sec_end_2 = time(&sec_end_2);
				
				fprintf(fw, "Solution size:%d\n", solution.size());
				fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());
				
				fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
				fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
				fprint_step_parallelism(fw, filtered_solution);
				
			//					print_fine_solution(filtered_solution);			
			
				fprintf(fw, "Solving time:%.3f,%d\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC, sec_end_1 - sec_begin_1);
				fprintf(fw, "Filtering time:%.3f,%d\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC, sec_end_2 - sec_begin_2);
			}
			else
			{
				break;
			}
			fflush(fw);
		}
		fclose(fw);
	}
}


void test_biconnected_shortening_BIG(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	clock_t begin_1, end_1;
	time_t sec_begin_1, sec_end_1;
	clock_t begin_2, end_2;
	time_t sec_begin_2, sec_end_2;

	int Seed;
	int init;
	int iter;
	int length;
	int limit;

	for (int round = 0; round <= 7; ++round)
	{
		switch (round)
		{
		case 0:
		{
			limit = 128;
			break;
		}
		case 1:
		{
			limit = 256;
			break;
		}
		case 2:
		{
			limit = 512;
			break;
		}
		case 3:
		{
			limit = 1024;
			break;
		}
		case 4:
		{
			limit = 2048;
			break;
		}
		case 5:
		{
			limit = 4096;
			break;
		}
		case 6:
		{
			limit = 8192;
			break;
		}
		case 7:
		{
			limit = 16384;
			break;
		}
		default:
		{
			break;
		}
		}

		for (length = 2; length <= 32; ++length)
		{			
			int ID = 0;
			Seed = 25;

			init = (length < 7) ? 7 : length;
			init = ((init % 2) == 0) ? init + 1 : init;

			iter = 1;
			int last_print = iter;

			printf("Estimating number of iterations... ");

			while (true)
			{
				GRAPH gg;

				srandom(Seed);
				generate_biconnected(init, iter, length, 2, &gg);

				if (2 * last_print < iter)
				{
					printf("%d ", iter);
					fflush(NULL);

					last_print = iter;
				}

				if (gg.m_vertices.size() > (unsigned int)limit)
				{
					break;
				}

				++iter;
			}
			printf("\n");

			char prefix[1024];
			char suffix[1024];
			char filename[1024];

			switch (method_type)
			{
			case BIBOX:
			{
				sprintf(prefix, "bibox_");
				break;
			}
			case BIBOX_MIT:
			{
				sprintf(prefix, "bibox_mit_");
				break;
			}
			case BIBOX_THETA_2:
			{
				sprintf(prefix, "bibox_theta_2_");
				break;
			}
			case BIBOX_THETA_3:
			{
				sprintf(prefix, "bibox_theta_3_");
				break;
			}
			case MIT:
			{
				sprintf(prefix, "mit_");
				break;
			}
			default:
			{
				break;
			}
			}
			
			switch (shortening_type)
			{
			case RAW:
			{
				sprintf(suffix, "biconnected_%s_%d-%d-%d^%d_raw.BIG.txt", SPECIFICATION, init, iter, length, limit);
				break;
			}
			case INVERSE:
			{
				sprintf(suffix, "biconnected_%s_%d-%d-%d^%d_inverse.BIG.txt", SPECIFICATION, init, iter, length, limit);
				break;
			}
			case REDUNDANT:
			{
				sprintf(suffix, "biconnected_%s_%d-%d-%d^%d_redundant.BIG.txt", SPECIFICATION, init, iter, length, limit);
				break;
			}
			case LONG:
			{
				sprintf(suffix, "biconnected_|%s|_%d-%d-%d^%d_long.BIG.txt", SPECIFICATION, init, iter, length, limit);
				break;
			}
			default:
			{
				break;
			}
			}
			
			FILE *fw;
			sprintf(filename, "%s%s", prefix, suffix);
			
			fw = fopen(filename, "w");
			
			if (fw == NULL)
			{
				printf("Cannot open %s. Terminating...\n", filename);
				exit(-1);
			}

			printf("Search for best seed... ");
			fflush(NULL);

			int best_Seed = Seed;
			int best_cycles = 0;
			
			for (int s = 0; s < MAX_SEED_TRIES; ++s)
			{
				GRAPH h;
				SOLUTION_DATABASE database;
				FINE_SOLUTION solution;

				srandom(Seed);
				generate_biconnected(init, iter, length, 2, &h);
				
				int cycles = 0;
				bool solved = solve_slider_obox_theta(&h, database, solution, false, false, cycles);
				
				if (solved)
				{
					best_Seed = Seed;
					break;
				}
				else
				{
					if (cycles > best_cycles)
					{
						printf("%d ", cycles);
						fflush(NULL);
						
						best_Seed = Seed;
						best_cycles = cycles;
					}
				}
				
				Seed = random();
			}
			
			printf("\n");
			Seed = best_Seed;
			
			int start_free;
			
			if (method_type == BIBOX)
			{
				start_free = 2;
			}
			else
			{
				start_free = 1;
			}
			
			printf("Solving for free... ");

			for (int free = start_free; /****/; ++free)
			{
				printf("%d (%d,%d,%d) ", free, DATABASE_HIT_COUNT, COMP_TOTAL_COUNT, COMP_SOLVED_COUNT);
				fflush(NULL);

				fprintf(fw, "ID:%d (%d,%d,%d)\n", ID, iter, length, free);
				++ID;
				
				SOLUTION_DATABASE database;
				FINE_SOLUTION solution, filtered_solution;
				
				MOVE_COUNT = 0;
				STEP_COUNT = 0;
				GRAPH g, h, ch, hv;		       
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &g);
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &h);
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &ch);
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &hv);
		
				fprintf(fw, "Graph |V|:%d\n", g.m_vertices.size());
		
				if (h.m_vertices.size() > (unsigned int)free)
				{
					begin_1 = clock();
					sec_begin_1 = time(&sec_begin_1);
					
					bool solved;
					
					switch (method_type)
					{
					case BIBOX:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
						break;
					}
					case BIBOX_MIT:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, true, false, true);
						break;
					}
					case BIBOX_THETA_2:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
						break;
					}
					case BIBOX_THETA_3:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
						break;
					}
					case MIT:
					{
						int solved_cycles = 0;
						solved = solve_slider_obox_theta(&h, database, solution, false, false, solved_cycles);
						break;
					}
					default:
					{
						solved = false;
						break;
					}
					}
// Bibox_theta 2
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
// Bibox_theta 3
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
// Bibox
//			bool solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
			
					end_1 = clock();
					sec_end_1 = time(&sec_end_1);
					begin_2 = clock();
					sec_begin_2 = time(&sec_begin_2);

					filter_fine_solution(&ch, solution, filtered_solution);
					
					switch (shortening_type)
					{
					case RAW:
					{
						break;
					}
					case INVERSE:
					{
						erase_trivial_inverses(filtered_solution);
						break;
					}
					case REDUNDANT:
					{
						erase_independent_inverses(filtered_solution);
						break;
					}
					case LONG:
					{
						shorten_solution(filtered_solution, &hv);
						break;
					}
					default:
					{
						break;
					}
					}

//		erase_trivial_inverses(filtered_solution);
//		erase_complex_inverses(filtered_solution);
//		erase_long_inverses(filtered_solution);
//		erase_independent_inverses(filtered_solution);
//		shorten_solution(filtered_solution, &hv);
			
					parallelize_fine_solution_multirobotic_FAST(filtered_solution);
				
					end_2 = clock();
					sec_end_2 = time(&sec_end_2);
					
					fprintf(fw, "Solution size:%d\n", solution.size());
					fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());

					fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
					fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
//					fprint_step_parallelism(fw, filtered_solution);

			//					print_fine_solution(filtered_solution);			

					fprintf(fw, "Solving time:%.3f,%d\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC, sec_end_1 - sec_begin_1);
					fprintf(fw, "Filtering time:%.3f,%d\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC, sec_end_2 - sec_begin_2);
				}
				else
				{
					break;
				}
				fflush(fw);
			}
			fclose(fw);
			++iter;
		}
	}
}


void test_biconnected_shortening_SCALE(METHOD_TYPE method_type = BIBOX_THETA_2, SHORTENING_TYPE shortening_type = RAW)
{
	clock_t begin_1, end_1;
	time_t sec_begin_1, sec_end_1;
	clock_t begin_2, end_2;
	time_t sec_begin_2, sec_end_2;

	int Seed;
	int init;
	int iter;
	int length;
	int Limit;
	int SFree = 2;

	for (int round = 1; round <= 7; ++round)
	{
		switch (round)
		{
		case 0:
		{
			Limit = 128;
			break;
		}
		case 1:
		{
			Limit = 256;
			break;
		}
		case 2:
		{
			Limit = 512;
			break;
		}
		case 3:
		{
			Limit = 1024;
			break;
		}
		case 4:
		{
			Limit = 2048;
			break;
		}
		case 5:
		{
			Limit = 4096;
			break;
		}
		case 6:
		{
			Limit = 8192;
			break;
		}
		case 7:
		{
			Limit = 16384;
			break;
		}
		default:
		{
			break;
		}
		}

		char prefix[1024];
		char suffix[1024];
		char filename[1024];

		switch (shortening_type)
		{
		case RAW:
		{
			sprintf(suffix, "biconnected_%s_%d-%d_raw.SCALE.txt", SPECIFICATION, Limit, SFree);
			break;
		}
		case INVERSE:
		{
			sprintf(suffix, "biconnected_%s_%d-%d_inverse.SCALE.txt", SPECIFICATION, Limit, SFree);
			break;
		}
		case REDUNDANT:
		{
			sprintf(suffix, "biconnected_%s_%d-%d_redundant.SCALE.txt", SPECIFICATION, Limit, SFree);
			break;
		}
		case LONG:
		{
			sprintf(suffix, "biconnected_%s_%d-%d_long.SCALE.txt", SPECIFICATION, Limit, SFree);
			break;
		}
		default:
		{
			suffix[0] = '\0';
			break;
		}
		}

		switch (method_type)
		{
		case BIBOX:
		{
			sprintf(prefix, "bibox_");
			break;
		}
		case BIBOX_MIT:
		{
			sprintf(prefix, "bibox_mit_");
			break;
		}
		case BIBOX_THETA_2:
		{
			sprintf(prefix, "bibox_theta_2_");
			break;
		}
		case BIBOX_THETA_3:
		{
			sprintf(prefix, "bibox_theta_3_");
			break;
		}
		case MIT:
		{
			sprintf(prefix, "mit_");
			break;
		}
		default:
		{
			prefix[0] = '\0';
			break;
		}
		}
		
		FILE *fw;
		sprintf(filename, "%s%s", prefix, suffix);
		
		fw = fopen(filename, "w");
			
		if (fw == NULL)
		{
			printf("Cannot open %s. Terminating...\n", filename);
			exit(-1);
		}

		int ID = 0;
		
		for (int limit = 235; limit <= Limit; ++limit)
		{
			for (length = 2; length <= 8; ++length)
			{			
				Seed = 25;
				
				init = (length < 7) ? 7 : length;
				init = ((init % 2) == 0) ? init + 1 : init;
				
				iter = 1;
				int last_print = iter;
				
				printf("Estimating number of iterations... ");
				
				while (true)
				{
					GRAPH gg;
					
					srandom(Seed);
					generate_biconnected(init, iter, length, 2, &gg);
					
					if (2 * last_print < iter)
					{
						printf("%d ", iter);
						fflush(NULL);
						
						last_print = iter;
					}
					
					if (gg.m_vertices.size() > (unsigned int)limit)
					{
						break;
					}
					
					++iter;
				}
				printf("\n");

				
				printf("Search for best seed... ");
				fflush(NULL);
				
				int best_Seed = Seed;
				int best_cycles = 0;
				
				for (int s = 0; s < MAX_SEED_TRIES; ++s)
				{
					GRAPH h;
					SOLUTION_DATABASE database;
					FINE_SOLUTION solution;
					
					srandom(Seed);
					generate_biconnected(init, iter, length, 2, &h);
					
					int cycles = 0;
					bool solved = solve_slider_obox_theta(&h, database, solution, false, false, cycles);
					
					if (solved)
					{
						best_Seed = Seed;
						break;
					}
					else
					{
						if (cycles > best_cycles)
						{
							printf("%d ", cycles);
							fflush(NULL);
							
							best_Seed = Seed;
							best_cycles = cycles;
						}
					}
				
					Seed = random();
				}
			
				printf("\n");
				Seed = best_Seed;
				
				int start_free = SFree;
				
				if (method_type == BIBOX)
				{
					start_free = 2;
				}
				else
				{
					start_free = 1;
				}
			
				printf("Solving for free... ");
				int free = start_free;
				
				printf("%d (%d,%d,%d) ", free, DATABASE_HIT_COUNT, COMP_TOTAL_COUNT, COMP_SOLVED_COUNT);
				fflush(NULL);
				
				fprintf(fw, "ID:%d (%d,%d,%d,%d)\n", ID, iter, length, free, limit);
				++ID;
				
				SOLUTION_DATABASE database;
				FINE_SOLUTION solution, filtered_solution;
				
				MOVE_COUNT = 0;
				STEP_COUNT = 0;
				GRAPH g, h, ch, hv;		       
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &g);
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &h);
					
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &ch);
				
				srandom(Seed);
				generate_biconnected(init, iter, length, free, &hv);
				
				fprintf(fw, "Graph |V|:%d\n", g.m_vertices.size());
				
				if (h.m_vertices.size() > (unsigned int)free)
				{
					begin_1 = clock();
					sec_begin_1 = time(&sec_begin_1);
					
					bool solved;
					
					switch (method_type)
					{
					case BIBOX:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
						break;
					}
					case BIBOX_MIT:
					{
							solved = solve_slider_bibox_theta(&h, database, solution, true, false, true);
							break;
					}
					case BIBOX_THETA_2:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
						break;
					}
					case BIBOX_THETA_3:
					{
						solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
						break;
					}
					case MIT:
					{
						int solved_cycles = 0;
						solved = solve_slider_obox_theta(&h, database, solution, false, false, solved_cycles);
						break;
					}
					default:
					{
						solved = false;
						break;
					}
					}
// Bibox_theta 2		
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, false, false);
// Bibox_theta 3
//			bool solved = solve_slider_bibox_theta(&h, database, solution, true, true, false);
// Bibox
//			bool solved = solve_slider_bibox_theta(&h, database, solution, false, false, false);
			
					end_1 = clock();
					sec_end_1 = time(&sec_end_1);
					begin_2 = clock();
					sec_begin_2 = time(&sec_begin_2);
					
					filter_fine_solution(&ch, solution, filtered_solution);
					
					switch (shortening_type)
					{
					case RAW:
					{
						break;
					}
					case INVERSE:
					{
						erase_trivial_inverses(filtered_solution);
						break;
					}
					case REDUNDANT:
					{
						erase_independent_inverses(filtered_solution);
						break;
					}
					case LONG:
					{
							shorten_solution(filtered_solution, &hv);
							break;
					}
					default:
					{
						break;
					}
					}
						
//		erase_trivial_inverses(filtered_solution);
//		erase_complex_inverses(filtered_solution);
//		erase_long_inverses(filtered_solution);
//		erase_independent_inverses(filtered_solution);
//		shorten_solution(filtered_solution, &hv);
			
					parallelize_fine_solution_multirobotic_FAST(filtered_solution);
					
					end_2 = clock();
					sec_end_2 = time(&sec_end_2);
					
					fprintf(fw, "Solution size:%d\n", solution.size());
					fprintf(fw, "Filtered solution size:%d\n", filtered_solution.size());
					
					fprintf(fw, "Parallel solution length:%d\n", calc_solution_duration(filtered_solution));
					fprintf(fw, "Parallelism:%.3f\n", calc_solution_parallelism(filtered_solution));
//					fprint_step_parallelism(fw, filtered_solution);

			//					print_fine_solution(filtered_solution);			
					
					fprintf(fw, "Solving time:%.3f,%d\n", (end_1 - begin_1) / (double)CLOCKS_PER_SEC, sec_end_1 - sec_begin_1);
					fprintf(fw, "Filtering time:%.3f,%d\n", (end_2 - begin_2) / (double)CLOCKS_PER_SEC, sec_end_2 - sec_begin_2);
				}
				else
				{
					break;
				}
				fflush(fw);
			}
		}
		fclose(fw);
		exit(0);
	}
}


int main(int argc, char **argv)
{
//	clock_t begin, end;

	printf("Mulirobot %s\n", VERSION);
	printf("(c) 2008-2010 %s\n\n", AUTHOR);

	SHORTENING_TYPE shortening_type;
	METHOD_TYPE method_type;
	
	if (argc == 5)
	{
		if (strcmp(argv[1], "mit") == 0)
		{
			method_type = MIT;
		}
		else if (strcmp(argv[1], "bibox") == 0)
		{
			method_type = BIBOX;
		}
		else if (strcmp(argv[1], "bibox_theta_2") == 0)
		{
			method_type = BIBOX_THETA_2;
		}
		else if (strcmp(argv[1], "bibox_theta_3") == 0)
		{
			method_type = BIBOX_THETA_3;
		}
		else if (strcmp(argv[1], "bibox_mit") == 0)
		{
			method_type = BIBOX_MIT;
		}
		else
		{
			printf("Wrong method type.\n");
			return -2;
		}
		
		if (strcmp(argv[2], "raw") == 0)
		{
			shortening_type = RAW;
		}
		else if (strcmp(argv[2], "inverse") == 0)
		{
			shortening_type = INVERSE;
		}
		else if (strcmp(argv[2], "redundant") == 0)
		{
			shortening_type = REDUNDANT;
		}
		else if (strcmp(argv[2], "long") == 0)
		{
			shortening_type = LONG;
		}
		else
		{
			printf("Wrong dataset.\n");
			return -4;
		}
		
		if (strcmp(argv[3], "grid") == 0 && strcmp(argv[4], "small") == 0)
		{
			test_grid_shortening(method_type, shortening_type);
		}
		else if (strcmp(argv[3], "biconnected") == 0 && strcmp(argv[4], "small") == 0)
		{
			test_biconnected_shortening(method_type, shortening_type);
		}
		else if (strcmp(argv[3], "grid") == 0 && strcmp(argv[4], "big") == 0)
		{
			test_grid_shortening_BIG(method_type, shortening_type);
		}
		else if (strcmp(argv[3], "biconnected") == 0 && strcmp(argv[4], "big") == 0)
		{
			test_biconnected_shortening_BIG(method_type, shortening_type);
		}
		else if (strcmp(argv[3], "grid") == 0 && strcmp(argv[4], "big") == 0)
		{
			printf("Wrong combination of parameters.\n");
			return -4;
		}
		else if (strcmp(argv[3], "biconnected") == 0 && strcmp(argv[4], "scale") == 0)
		{
			test_biconnected_shortening_SCALE(method_type, shortening_type);
		}
		else
		{
			printf("Wrong dataset.\n");
			return -3;
		}
	}
	else
	{
		printf("Wrong number of parameters.\n");
		return -1;
	}

	/*
	test_exchange_grids();

	test_grid_shortening(BIBOX_THETA_2, RAW);
	test_grid_shortening(BIBOX_THETA_2, INVERSE);
	test_grid_shortening(BIBOX_THETA_2, REDUNDANT);
	test_grid_shortening(BIBOX_THETA_2, LONG);

	test_grid_shortening(MIT, RAW);
	test_grid_shortening(MIT, INVERSE);
	test_grid_shortening(MIT, REDUNDANT);
	test_grid_shortening(MIT, LONG);

	test_biconnected_shortening(BIBOX_THETA_2, RAW);
	test_biconnected_shortening(BIBOX_THETA_2, INVERSE);
	test_biconnected_shortening(BIBOX_THETA_2, REDUNDANT);
	test_biconnected_shortening(BIBOX_THETA_2, LONG);

	test_biconnected_shortening(MIT, RAW);
	test_biconnected_shortening(MIT, INVERSE);
	test_biconnected_shortening(MIT, REDUNDANT);
	test_biconnected_shortening(MIT, LONG);
	*/
}
